/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>

#ifdef __MWERKS__
#include <console.h>
#endif	

#include "lik.h"
#include "calclik.h"
#include "maxlard.h"
#include "triplet.h"
#include "fF84.h"
#include "gamma.h"
#include "random.h"

#define PROGRAM_NAME "lard"
#define VERSION_NUMBER "Version 2.2"

char *ModelNames[numModels]={
	"F84",
	"HKY",
	"REV"
};

int model, printHelp;

/* prototypes */

static void PrintTitle();
static void PrintUsage();
void PrintVerbose();
void ReadParams();
void ReadFileParams();
void ReadFile();

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Likelihood Analysis of Recombination in DNA - LARD\n");
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 1999 Andrew Rambaut\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{
	fprintf(stderr, "Usage: %s [-mMODEL] [-r #] [-e[r] | -d] [-w] [-s #] [-z #] [-u #1 [#2]] [-p nnn] [-c #1 #2 [#3]] [-g #] [-a#] [-n #] [-h|-help] [parameters] [-v[m][p]|-q]\n", PROGRAM_NAME);
	fprintf(stderr, "  -m: MODEL = F84, HKY or REV [default=HKY]\n");
	fprintf(stderr, "  -r: # = number of recombination points [default=1]\n");
	fprintf(stderr, "  -e: use standard heuristic search [default=exhaustive]\n");
	fprintf(stderr, "    r Random starting position\n");
	fprintf(stderr, "  -d: use great deluge heuristic search [default=exhaustive]\n");
	fprintf(stderr, "  -w: write likelihood surface (exhaustive search only)\n");
	fprintf(stderr, "  -s: # = recombination point step [default=1, 3 for codons]\n");
	fprintf(stderr, "  -z: # = recombination region size [default=3]\n");
	fprintf(stderr, "  -u: #1 [ #2] = user recombination points [default=search]\n");
	fprintf(stderr, "  -p: CODON CATEGORIES = 112, 123, 120, etc. [homogeneity]\n");
	fprintf(stderr, "  -c: #1 #2 #3 = codon category rates [default = homogeneity].\n");
	fprintf(stderr, "  -g: # = num categories for gamma rate heterogeneity [default = homogeneity].\n");
	fprintf(stderr, "  -a: # = gamma shape (alpha) for site rate heterogeneity [default = 0.5].\n");
	fprintf(stderr, "  -n: # = Number of Datasets [default=1]\n");
	fprintf(stderr, "  -h,-help = Give this help message (with details of selected model)\n");
	fprintf(stderr, "  parameters = Specific to model.  To get details use: %s -mMODEL -help \n", PROGRAM_NAME);
	fprintf(stderr, "  -v: Verbose information\n");
	fprintf(stderr, "    m Memory usage information\n");
	fprintf(stderr, "    p Show indication of progress\n");
	fprintf(stderr, "  -q: Quiet\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j, x, stepDefined;
	char ch;
	char *P, st[4];
	int r[2];
	
	model=HKY;
	printHelp=0;
	numDatasets=1;
	
	numModelCats=1;

	verbose=0;
	verboseMemory=0;
	showProgress=0;
	quiet=0;

	stepDefined=0;
	numCross=1;
	minRegionStep=1;
	minRegionSize=3;
	heuristic=0;
	greatDeluge=0;
	randomStart=0;
	writeSurface=0;

	userRecomb=0;
	userPoint1=0;
	userPoint2=0;
	
	debug=0;
	
	numCategories=1;
	codonCats[0]=codonCats[1]=codonCats[2]=0;
	for (i=0; i<MAX_CATEGORIES; i++)
		siteRate[i]=1.0;
	codonRates=0;
	estimateRates=0;
	gammaRates=0;
	alpha=1.0;
	estimateAlpha=0;
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (P[0]=='-') {
			ch=toupper(P[1]);
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'Q':
					quiet=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P+2, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'R':
					if (GetIntParams(argc, argv, &i, P+2, 1, &numCross) || numCross<1 || numCross>2) {
						fprintf(stderr, "Bad number of recombination points: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}

	if (!quiet || printHelp)
 		PrintTitle();
	if (printHelp)
 		PrintUsage();
	SetModel(printHelp);
	if (printHelp)
		exit(0);
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (*P!='-' && *P!='+') {
			fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
			PrintUsage();
			exit(0);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'M': 	/* although this has already been read, we need to
							   use it up in this run through of the params.    */
					GetStrParam(argc, argv, &i, P, st, 3);
				break;
				case 'H':
				case 'Q':
				case 'R':
					GetIntParams(argc, argv, &i, P, 1, &x);
				break;
				case 'E':
					heuristic=1;
					while (*P!='\0') {
						switch (*P) {
							case 'R': randomStart=1; break;
							default:					
								fprintf(stderr, "Unknown heuristic mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
					if (greatDeluge) {
						fprintf(stderr, "Please choose -e (heuristic) or -d (great deluge) but not both.\n\n");
						PrintUsage();
						exit(0);
					}
				break;
				case 'D':
					greatDeluge=1;
					if (heuristic) {
						fprintf(stderr, "Please choose -e (heuristic) or -d (great deluge) but not both.\n\n");
						PrintUsage();
						exit(0);
					}
				break;
				case 'W':
					writeSurface=1;
				break;
				case 'S':
					if (GetIntParams(argc, argv, &i, P, 1, &minRegionStep) || minRegionStep<1) {
						fprintf(stderr, "Bad minimum region step: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					stepDefined=1;
				break;
				case 'Z':
					if (GetIntParams(argc, argv, &i, P, 1, &minRegionSize) || minRegionSize<1) {
						fprintf(stderr, "Bad minimum region size: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'U':
					if (GetIntParams(argc, argv, &i, P, numCross, r)) {
						fprintf(stderr, "Bad user recombination point(s): %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					userRecomb=1;
					userPoint1=r[0];
					if (numCross==2)
						userPoint2=r[1];
				break;
				case 'P':
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Codon Categories Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					P=st;
					if (strlen(P)==3) {
						int cp1=0, cp2=0, cp3=0;
						
						for (j=0; j<3; j++) {
							codonCats[j]=*P-'1';
							switch (codonCats[j]) {
								case -1: break;  /* exclude this codon position */
								case 0: cp1=1; break;
								case 1: cp2=1; break;
								case 2: cp3=1; break;
								default:
									fprintf(stderr, "Codon categories should be from 0-3: %s\n\n", argv[i]);
									PrintUsage();
									exit(0);
							}
							P++;
						}
						if (!cp1 || (cp3 && !cp2)) {
							fprintf(stderr, "Codon categories should be from 0-3: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						numCategories=cp1+cp2+cp3;
						codonRates=1;
					} else {
						fprintf(stderr, "Must have a category for each codon position: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'C':
					estimateRates=0;
					if (numCategories==3) {
						if (GetDoubleParams(argc, argv, &i, P, 3, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rates: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else if (numCategories==2) {
						if (GetDoubleParams(argc, argv, &i, P, 2, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rate: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else {
						fprintf(stderr, "Must have two or three codon categories set to specify rates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'G':
					if (GetIntParams(argc, argv, &i, P, 1, &numCategories) || numCategories<1 || numCategories>MAX_CATEGORIES) {
						fprintf(stderr, "Bad number of Gamma rate categories: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					gammaRates=1;
				break;
				case 'A':
					if (gammaRates) {
						if (GetDoubleParams(argc, argv, &i, P, 1, &alpha) || alpha<0.0) {
							fprintf(stderr, "Bad alpha shape for Gamma rate heterogeneity: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						estimateAlpha=0;
					} else {
						fprintf(stderr, "Must specify the number of gamma rate categories before setting alpha: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
						fprintf(stderr, "Bad number of Datasets: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'V':
					verbose=1;
					while (*P!='\0') {
						switch (*P) {
							case 'P': showProgress=1; break;
							case 'M': verboseMemory=1; break;
							default:					
								fprintf(stderr, "Unknown verbose mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				default:
					if (ReadModelParam(argc, argv, &i)==-1) {
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}
	
	if (numCategories<2) {
		estimateRates=0;
		estimateAlpha=0;
		gammaRates=0;
	}
	
	if (gammaRates)
		estimateRates=0;
		
	if (codonRates && !stepDefined)
		minRegionStep=3;
			
	if (FinishModelParam()<0) {
		PrintUsage();
		exit(0);
	}
}


void ReadFileParams()
{
	char ch, st[256];
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
	ungetc(ch, stdin);
	fgets(st, 255, stdin);
	if ( sscanf( st, " %d %d", &numTaxa, &numBases)!=2 ) {
		fprintf(stderr, "Unable to read parameters from standard input\n");
		exit(0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numBases) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numBases ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numTaxa && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numTaxa );
}


void PrintVerbose()
{
	int i;
	
	fprintf(stderr, "%d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "	%d distinct patterns\n", numPatterns);
	if (numCategories>1) {
		fprintf(stderr, "Rate heterogeneity:\n");
		if (gammaRates) {
			fprintf(stderr, "  Discrete Gamma rate heterogeneity with %d categories\n", numCategories);
			fprintf(stderr, "  shape parameter, alpha = %f\n", alpha);
		} else {
			fprintf(stderr, "  Number of codon position categories = %d, pattern = %d%d%d\n", 
							numCategories, codonCats[0]+1, codonCats[1]+1, codonCats[2]+1);
			fprintf(stderr, "  rates for categories supplied by user\n");
		}
		fprintf(stderr, "    rates =");
		for (i=0; i<numCategories; i++)
			fprintf(stderr, " %f", siteRate[i]);
		fputc('\n', stderr);
	} else
		fprintf(stderr, "Rate homogeneity assumed");
	fprintf(stderr, ".\n");
	fprintf(stderr, "Model=%s\n", ModelNames[model]);
}
	
	
int main(int argc, char **argv)
{
	int i, n, oldTaxa, oldBases;
	double lik0, lik1;
	TTree *tree;
	clock_t coreStart, totalStart;
	double coreSecs, totalSecs;
#if __profile__
	OSErr  theErr;
#endif
	
#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif	

	totalMem=0;
 	
	totalStart = clock();
	
	SetSeed(time(NULL));
	
	ReadParams(argc, argv);

	ReadFileParams();
	if (numTaxa!=3) {
		fprintf(stderr, "The input dataset must consist of three sequences\n");
		exit(0);
	}
	
	AllocateMemory();
		
	ReadFile();
	
	if (gammaRates)
		DiscreteGamma(freqRate, siteRate, alpha, alpha, numCategories, 0);

	CreatePatterns();

	tree=CreateTriplet();
		
	numSets=numDatasets;
			
	if (!quiet && verbose)
 		PrintVerbose(tree);

	AllocateTreeLikelihoods(tree);
	SetTipLikelihoods(tree);
	SetupLikelihood(tree);
	SetTriplet(tree);
	 
	if (writeSurface) {
		if ((surfaceFile=fopen("likelihood.surface", "wt"))==NULL) {
			fprintf(stderr, "Failed to open likelihood surface file for writing\n");
			exit(0);
		}
	} else
		surfaceFile=NULL;
		
	coreStart = clock();

	if (numSets>1) {
		if (!quiet) {
			fprintf(stdout, "no.\tlnL(H0)\tlnL(H1)\tdiff");
			if (numCross==1)
				fprintf(stdout, "\tcross");
			else
				fprintf(stdout, "\tfrom\tto");
			fprintf(stdout, "\n");
			fflush(stdout);
		}
		
		for (n=0; n<numSets; n++) {
			LardLik(tree);
							
			fprintf(stdout, "%d\t%f\t%f\t%f\t%d", n+1, tree->lnL, tree->lnLRecomb, tree->lnL-tree->lnLRecomb, tree->recomb1);
			if (numCross>1)
				fprintf(stdout, "\t%d", tree->recomb2);
			fprintf(stdout, "\n");
			fflush(stdout);
			
			if (n<numDatasets-1) {
				oldTaxa=numTaxa; oldBases=numBases;
				ReadFileParams();
				if (oldTaxa!=numTaxa || oldBases!=numBases) {
					fprintf(stderr, "The number of taxa and bases in each dataset must be the same.\n");
					exit(0);
				}       
				ReadFile();
				CreatePatterns();
				SetTipLikelihoods(tree);
				SetupLikelihood(tree);
			}
		}
	} else {
		LardLik(tree);
	}
	coreSecs = (double)(clock() - coreStart) / CLOCKS_PER_SEC;		
	
	if (writeSurface)
		fclose(surfaceFile);

	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	
	if (!quiet) {
		fprintf(stderr, "\nTime taken - core: %G, total: %G seconds\n", coreSecs, totalSecs);
	}
	if (!quiet && verboseMemory)
		fprintf(stderr, "Total memory used: %ld\n", totalMem);
		
	return 0;
}



