/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lik.h"
#include "triplet.h"
#include "calclik.h"
#include "maxlard.h"
#include "gamma.h"
#include "search.h"
#include "progress.h"
#include "random.h"

#define MAX_RESTARTS 5

static int currBranch, currLUT, currRecomb1, currRecomb2, currRegion;
static TTree *currTree;
static double likRegion[2];
static int numSteps, maxRecomb1, maxRecomb2;

/* prototypes */
double ExhaustivePoint();
double HeuristicRegion();
double ExhaustiveRegion();

double Score(int x, int y);
double LineSearch(int x, int y, int *newX, int *newY);
int Peturb(int oldX, int oldY, int bestX, int bestY);
double GreatDeluge(double inBound, double w);

int CalcEstRem(double oldDiff, double newDiff);
double MaxBranchLengths(double tol);
double MaxBranchLengthsNoRecomb(double tol);
double BranchChangedLik(double);

int CalcEstRem(double oldDiff, double newDiff)
{
	double ratio, diff;
	int i=0;
	
	ratio=newDiff/oldDiff;
	diff=newDiff;
	
	while (diff>STOL) {
		diff*=ratio;
		i++;
	}
	
	return i;
}

/*********************/
void LardLik(TTree *tree)
{	
	double lik, maxLik;
	int i, cat;
		
	numSteps=numBases/minRegionStep;
	if (numSteps<2) {
		fprintf(stderr, "The recombination point step size is too big.");
		exit(0);
	}
	
	if (!quiet && verbose) {
		if (numCross==1)
			fprintf(stderr, "  1 recombination point\n");
		else
			fprintf(stderr, "  2 recombination point\n");
		
		if (userRecomb) {
			if (numCross==1)
				fprintf(stderr, "  User defined recombination point : %d\n\n", userPoint1);
			else
				fprintf(stderr, "  User defined recombination points : %d and %d\n\n", userPoint1, userPoint2);
		} else
			fprintf(stderr, "  Recombination point step size: %d\n\n", minRegionStep);
	}

	for (i=0; i<numBranches; i++)
		branchValue[i]=0.1;
		
	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);

	startNode=tree->start;
	currTree=tree;
	
	currRecomb1=numBases;
	currRecomb2=numBases;

	lik=MaxBranchLengthsNoRecomb(FTOL2);		
	tree->lnL=lik;

	if (!quiet && numDatasets==1) {
		fprintf(stderr, "No recombination: lik=%f\n", lik);
		fprintf(stderr, "Branches: %s: %f\t%s:%f\t%s:%f\n\n", 
					names[tree->nodes[0].tipNo], branchValue[tree->nodes[0].branchNo],
					names[tree->nodes[1].tipNo], branchValue[tree->nodes[1].branchNo],
					names[tree->nodes[2].tipNo], branchValue[tree->nodes[2].branchNo]);
	}

	if (userRecomb) {
		maxRecomb1=userPoint1;
		maxRecomb2=userPoint2;
	} else {
		if (numCross==1)
			ExhaustivePoint();
		else if (heuristic)
			HeuristicRegion();
		else if (greatDeluge)
			GreatDeluge(lik*2, 0.05);
		else
			ExhaustiveRegion();
	}
	
	if (maxRecomb1<maxRecomb2) {
		currRecomb1=maxRecomb1;
		currRecomb2=maxRecomb2;
	} else {
		currRecomb1=maxRecomb2;
		currRecomb2=maxRecomb1;
	}
	maxLik=MaxBranchLengths(FTOL2);	

	if (!quiet && numDatasets==1) {
		if (numCross==1)
			fprintf(stderr, "Recombination point (to the left of): %d\tlik=%f LR=%f\n", currRecomb1+1, maxLik, lik-maxLik);
		else
			fprintf(stderr, "Recombination region (to the left of): %d -> %d\tlik=%f LR=%f\n", currRecomb1+1, currRecomb2+1, maxLik, lik-maxLik);
		fprintf(stderr, "First region: %s: %f\t%s:%f\t%s:%f\n", 
					names[tree->nodes[0].tipNo], branchValue[tree->nodes[0].branchNo],
					names[tree->nodes[1].tipNo], branchValue[tree->nodes[1].branchNo],
					names[tree->nodes[2].tipNo], branchValue[tree->nodes[2].branchNo]);
		fprintf(stderr, "Second region: %s: %f\t%s:%f\t%s:%f\n", 
					names[tree->nodes[0].tipNo], branchValue[tree->nodes[0].branchNo+3],
					names[tree->nodes[1].tipNo], branchValue[tree->nodes[1].branchNo+3],
					names[tree->nodes[2].tipNo], branchValue[tree->nodes[2].branchNo+3]);
	}
				
	tree->lnLRecomb=maxLik;
	tree->recomb1=maxRecomb1+1;
	tree->recomb2=maxRecomb2+1;
}

/*********************/
double ExhaustivePoint()
{	
	double lik, maxLik;
	
	maxRecomb2=numBases;

	if (!quiet && showProgress)		
		InitProgressBar(numSteps);

	if (!quiet && showProgress)
		DrawProgressBar();
		
	maxLik=BIG;
	currRecomb2=numBases;
	for (currRecomb1=minRegionStep; currRecomb1<numBases-1; currRecomb1+=minRegionStep) {
		lik=MaxBranchLengths(FTOL);		
		if (lik<maxLik) {
			maxLik=lik;
			maxRecomb1=currRecomb1;
		}
		
		if (writeSurface)
			fprintf(surfaceFile, "%d\t%lf\n", currRecomb1, lik);
			
		if (!quiet && showProgress)
			ProgressBar();
	}

	return maxLik;
}

/*********************/
/*double HeuristicRegion()
{	
	double lik, maxLik, lastLik, diff, oldDiff;
	int i, recomb1, recomb2, estRem;
	
	if (!quiet && showProgress)		
		InitProgressBar(numSteps);

	lastLik=lik=BIG;
	
	if (randomStart)
		maxRecomb2=(int)floor(((double)numBases)*rndu());
	else
		maxRecomb2=numBases;
		
	i=0;
	do {
		if (!quiet && showProgress)
			DrawProgressBar();
		maxLik=BIG;
		recomb2=maxRecomb2;
		for (recomb1=minRegionStep; recomb1<numBases-1; recomb1+=minRegionStep) {
			if (recomb1<recomb2) {
				currRecomb1=recomb1;
				currRecomb2=recomb2;
			} else {		
				currRecomb1=recomb2;
				currRecomb2=recomb1;
			}		

			lik=MaxBranchLengths(FTOL);		
			if (lik<maxLik) {
				maxLik=lik;
				maxRecomb1=currRecomb1;
				maxRecomb2=currRecomb2;
			}
			if (!quiet && showProgress)
				ProgressBar();
		}
		if (numCross!=1) {
			if (!quiet && showProgress)
				DrawProgressBar();
			if (maxLik!=lastLik) {
				recomb1=maxRecomb1;
				for (recomb2=numBases-minRegionStep; recomb2>0; recomb2-=minRegionStep) {
					if (recomb1<recomb2) {
						currRecomb1=recomb1;
						currRecomb2=recomb2;
					} else {		
						currRecomb1=recomb2;
						currRecomb2=recomb1;
					}		
					
					lik=MaxBranchLengths(FTOL);		
					if (lik<maxLik) {
						maxLik=lik;
						maxRecomb1=currRecomb1;
						maxRecomb2=currRecomb2;
					}
					if (!quiet && showProgress)
						ProgressBar();
				}
			}
				
			oldDiff=diff;				
			diff=lastLik-maxLik;
			if (diff<-EPS)
				diff=-diff;
			lastLik=maxLik;
			i++;

			if (!quiet && showProgress) {
				if (i>2) {
					estRem=CalcEstRem(oldDiff, diff);
					fprintf(stderr, "%d/%d: %-10.4f", i, i+estRem, maxLik);
				} else
					fprintf(stderr, "%d: %-10.4f", i, maxLik);
				if (diff>1.0) 
					fprintf(stderr, " >1.0 (%d->%d)\n\n", maxRecomb1, maxRecomb2);
				else
					fprintf(stderr, " %-8.4f (%d->%d)\n\n", diff, maxRecomb1, maxRecomb2);
			}
		} else
			diff=0.0;
	} while (diff>STOL);

	return maxLik;
}*/

/*********************/
double HeuristicRegion()
{	
	double lik, maxLik, bestLik;
	int i, j, recomb1, recomb2, bestRecomb1, bestRecomb2, converged;
	
	maxLik=BIG;
	
	for (j=0; j<3; j++) {
		bestLik=BIG;
		do {
			recomb1=(int)floor(((double)((numBases+1)/minRegionStep))*rndu())*minRegionStep;
			recomb2=(int)floor(((double)((numBases+1)/minRegionStep))*rndu())*minRegionStep;
		} while (recomb1==recomb2);
		
		if (recomb1<recomb2) {
			bestRecomb1=recomb1;
			bestRecomb2=recomb2;
		} else {
			bestRecomb1=recomb2;
			bestRecomb2=recomb1;
		}

		converged=0;
		
		i=0;
		while (!converged) {
			lik=LineSearch(bestRecomb1, bestRecomb2, &recomb1, &recomb2);
			if ((bestRecomb1==recomb1 && bestRecomb2==recomb2) ||
				(bestRecomb1==recomb2 && bestRecomb2==recomb1)) {
				converged=1;
			} else if (lik<bestLik) {
				if (recomb1<recomb2) {
					bestRecomb1=recomb1;
					bestRecomb2=recomb2;
				} else {
					bestRecomb1=recomb2;
					bestRecomb2=recomb1;
				}
				bestLik=lik;

				if (bestLik<maxLik) {
					maxRecomb1=bestRecomb1;
					maxRecomb2=bestRecomb2;
					maxLik=bestLik;
				}
			}

			i++;

			if (!quiet && showProgress)
				fprintf(stderr, "Random start %d, iteration %d: %-10.4f (%d->%d)\n\n", 
							j+1, i, bestLik, bestRecomb1, bestRecomb2);
		} 
		if (!quiet && showProgress)
			fprintf(stderr, "Best: %-10.4f (%d->%d)\n\n", maxLik, maxRecomb1, maxRecomb2);
	}
			
	return maxLik;
}

/*********************/
double Score(int x, int y)
{	
	double lik;
	
	if (x<y) {
		currRecomb1=x;
		currRecomb2=y;
	} else {
		currRecomb1=y;
		currRecomb2=x;
	}
	
	if (currRecomb1<0)
		currRecomb1=0;
	if (currRecomb2>numBases)
		currRecomb2=numBases;
	
	lik=MaxBranchLengths(FTOL);	
	
	return lik;
}

/*********************/
double LineSearch(int x, int y, int *newX, int *newY)
{	
	double lik, maxLik;
	int recomb1, recomb2, maxX, maxY;
	
	if (!quiet && showProgress)	{	
		InitProgressBar((numBases + abs(x-y))/minRegionStep);
		DrawProgressBar();
	}
		
	maxLik=BIG;
	
	maxX=0;
	maxY=x+minRegionStep;
	
	for (recomb1=0; recomb1<y; recomb1+=minRegionStep) {
		lik=Score(recomb1, y);		
		if (lik<maxLik) {
			maxLik=lik;
			maxX=recomb1;
			*newX=maxX;
			*newY=y;
		}
		if (!quiet && showProgress)
			ProgressBar();
	}

	for (recomb2=x+minRegionStep; recomb2<=numBases; recomb2+=minRegionStep) {
		lik=Score(x, recomb2);		
		if (lik<maxLik) {
			maxLik=lik;
			maxY=recomb2;
			*newX=x;
			*newY=maxY;
		}
		if (!quiet && showProgress)
			ProgressBar();
	}

	lik=Score(maxX, maxY);		
	if (lik<maxLik) {
		maxLik=lik;
		*newX=maxX;
		*newY=maxY;
	}

	return maxLik;
}

/*********************/
int Peturb(int oldX, int oldY, int bestX, int bestY)
{	
	double r;
	int jump, newX;
	
	do {
		if (rndu()<0.05) { /* Select one of the existing best break points */
			if (rndu()>=0.5)
				newX=bestX;
			else
				newX=bestY;
		} else {
			/* Draw from an inverse square but then scale it and
			   adjust it to start from the origin */
			r=(sqrt(1/ rndu())*25) - 25;
			if (rndu()>=0.5) {	/* Add the peturbation */
				newX = oldX + (int)r;
				
				while (newX>numBases)
					newX-=numBases;

			} else {			/* Subtract */
				newX = oldX - (int)r;
				
				while (newX<0)
					newX+=numBases;
					
			}
		}
	} while (newX==oldX || newX==oldY);
	
	return newX;

}

/*int Peturb(int oldX, int oldY, int bestX, int bestY)
{	
	double r, r1;
	int newX;
	
	do {
		r1=rndu();
		if (r1<0.05) {
			if (rndu()>=0.5)
				newX=bestX;
			else
				newX=bestY;
		} if (r1<0.25) {
			newX=(int)floor(((double)numBases)*rndu());
		} else {
			if (r1<0.5)
				r=rndu()*10;
			else
				r=rndu()*100;
							
			if (rndu()>=0.5)
				r*=-1.0;
			
			newX+=r;
			if (newX>numBases)
				newX=numBases;
			if (newX<0)
				newX=0;
		}
	} while (newX==oldX || newX==oldY);
	
	return newX;

}*/

/*********************/
double GreatDeluge(double inBound, double w)
{	
	/* possible tabu search */
	
	double lik, lik1, bestLik, maxLik, bound;
	int i, j, sx, sy, sx1, sy1, bestsx, bestsy, lastLineX, lastLineY;
	int found, finish;
	
	lastLineX=lastLineY=-1;
	maxLik=inBound;
	
	finish=0;
	i=0;
	do { /* loop through random starts */
		bound = bestLik = inBound;
		
		/* random start position */
		do {
			sx=(int)floor(((double)numBases+1)*rndu());
			sy=(int)floor(((double)numBases+1)*rndu());
		} while (sx==sy);
		
		lik=Score(sx, sy);	

		do { /* loop through great deluge paces */
			j=0;	
			found = 0;
			do { /* loop until acceptable solution found */
			
				/* perturb s->s1 */
				if (rndu()>=0.5) {
					sy1=sy;
					sx1=Peturb(sx, sy, bestsx, bestsy);
				} else {
					sx1=sx;
					sy1=Peturb(sy, sx, bestsx, bestsy);
				}

				lik1=Score(sx1, sy1);	

				if (lik1<bound) {
					lik=lik1;

					sx=sx1;
					sy=sy1;
					
					if (lik1<bestLik) {
						bestLik=lik1;
						if (sx1<sy1) {
							bestsx=sx1;
							bestsy=sy1;
						} else {
							bestsx=sy1;
							bestsy=sx1;
						}

						if (!quiet && showProgress) {
							fprintf(stderr, "bound: %-8.4f\tS: %-8.4f (%d->%d)\tbest: %-8.4f (%d->%d)\n", bound, lik, sx, sy, bestLik, bestsx, bestsy);
						}
				
					}
					
					bound -= (bound - lik1) * w;
					
					found = 1;
				}
				
				j++;
			} while (j<20 && !found);
			
			if (!found) {
				if ( !(lastLineX==bestsx && lastLineY==bestsy) ) {
				
					lik1=LineSearch(bestsx, bestsy, &sx1, &sy1);
					if (lik1<bestLik) {
						bestLik=lik1;
						if (sx1<sy1) {
							bestsx=sx1;
							bestsy=sy1;
						} else {
							bestsx=sy1;
							bestsy=sx1;
						}
						found=1;
					}
					lastLineX=bestsx;
					lastLineY=bestsy;
					if (found && !quiet && showProgress) {
						fprintf(stderr, "line search: %-8.4f (%d->%d)\n", bestLik, bestsx, bestsy);
					}
				}
			}
								
		} while (found);
		
		if (!quiet && showProgress) {
			fprintf(stderr, "random start %d, likelihood: %-8.4f (%d->%d)\n\n", i+1, bestLik, bestsx, bestsy);
		}
	
		if (maxLik==bestLik) 
			finish=1;
			
		if (bestLik<maxLik) {
			maxLik=bestLik;
			maxRecomb1=bestsx;
			maxRecomb2=bestsy;
		}
		
		i++;
		
	} while (!finish && i<MAX_RESTARTS);


	return bestLik;
}

/*********************/
double ExhaustiveRegion()
{	
	double lik, maxLik;
	int recomb1, recomb2;
	
	if (!quiet && showProgress)		
		InitProgressBar(numBases-minRegionSize-minRegionSize);

	if (!quiet && showProgress)
		DrawProgressBar();
	maxLik=BIG;
/*	for (recomb1=minRegionSize; recomb1<numBases-minRegionSize; recomb1+=minRegionStep) {
		if (writeSurface)
			fprintf(surfaceFile, "%d", recomb1);
			
		for (recomb2=0; recomb2<numBases-recomb1; recomb2+=minRegionStep) {
			currRecomb1=recomb2;
			currRecomb2=recomb2+recomb1;

			lik=MaxBranchLengths(FTOL);		
			if (lik<maxLik) {
				maxLik=lik;
				maxRecomb1=currRecomb1;
				maxRecomb2=currRecomb2;
			}

			if (writeSurface)
				fprintf(surfaceFile, "\t%lf", lik);
		}
		
//		fprintf(stderr, "size: %d\tmax pos: %d\tlik: %lf", recomb1, recomb2, maxLik);

		if (!quiet && showProgress)
			ProgressBar();
			
		if (writeSurface)
			fprintf(surfaceFile, "\n");
	}
*/
/*	if (writeSurface) {
		for (recomb2=0; recomb2<=numBases; recomb2+=minRegionStep)
			fprintf(surfaceFile, "\t%d", recomb2);
		fprintf(surfaceFile, "\n");
	}
	
	for (recomb1=0; recomb1<numBases; recomb1+=minRegionStep) {
		if (writeSurface) {
			fprintf(surfaceFile, "%d", recomb1);
			for (recomb2=0; recomb2<=recomb1; recomb2+=minRegionStep)
				fprintf(surfaceFile, "\t.");
		}
			
		for (recomb2=recomb1+minRegionSize; recomb2<=numBases; recomb2+=minRegionStep) {
			currRecomb1=recomb1;
			currRecomb2=recomb2;

			lik=MaxBranchLengths(FTOL);		
			if (lik<maxLik) {
				maxLik=lik;
				maxRecomb1=currRecomb1;
				maxRecomb2=currRecomb2;
			}

			if (writeSurface)
				fprintf(surfaceFile, "\t%lf", lik);
		}
		
		if (!quiet && showProgress)
			ProgressBar();
			
		if (writeSurface)
			fprintf(surfaceFile, "\n");
	}
*/
	for (recomb1=0; recomb1<numBases; recomb1+=minRegionStep) {			
		for (recomb2=recomb1+minRegionSize; recomb2<=numBases; recomb2+=minRegionStep) {
			currRecomb1=recomb1;
			currRecomb2=recomb2;

			lik=MaxBranchLengths(FTOL);		
			if (lik<maxLik) {
				maxLik=lik;
				maxRecomb1=currRecomb1;
				maxRecomb2=currRecomb2;
			}

			if (writeSurface)
				fprintf(surfaceFile, "%d\t%d\t%lf\n", recomb1, recomb2, lik);
		}
		
		if (!quiet && showProgress)
			ProgressBar();
	}

	return maxLik;
}

/*************************************/
double MaxBranchLengths(double tol)
{
	int n, i;
	double lik, lastLik, dummy;
	
	likRegion[0]=likRegion[1]=0.0;
	lik=1E30;
	
	n=0;
	do { 
		startNode=startNode->next;
	
		for (i=0; i<numBranches; i++) {
			currBranch=i;
			currRegion=i/3;
			currLUT=i << 4;
			likRegion[currRegion]=brent(EPS, branchValue[i], MAX_BRANCH_LENGTH_VAL, BranchChangedLik, tol, &dummy);
/*			branchValue[i]=dummy;
			for (cat=0; cat<numCategories; cat++)
				MatrixFunc(transMatrix[cat]+currLUT, cat, currBranch);*/
		}
		
		lastLik=lik;
		lik=likRegion[0]+likRegion[1];
		n++;
	} while (lastLik-lik>tol);
	
	return lik;
}

/*************************************/
double MaxBranchLengthsNoRecomb(double tol)
{
	int n, i;
	double lik, lastLik, dummy;
	
	startNode=startNode->next;
	lik=1E30;

	n=0;
	do { 
		lastLik=lik;
		for (i=0; i<numBranches; i++) {
			currBranch=i;
			currRegion=0;
			currLUT=i << 4;
			lik=brent(EPS, branchValue[i], MAX_BRANCH_LENGTH_VAL, BranchChangedLik, tol, &dummy);
	/*		branchValue[i]=dummy;
			for (cat=0; cat<numCategories; cat++)
				MatrixFunc(transMatrix[cat]+currLUT, cat, currBranch);*/
		}
		
		n++;
	} while (lastLik-lik>tol);
	
	return lik;
}

/*************************************/
double BranchChangedLik(double value)
{
	int cat;
	double lik;

	branchValue[currBranch]=value;

	for (cat=0; cat<numCategories; cat++)
		MatrixFunc(transMatrix[cat]+currLUT, cat, currBranch);

	lik=CalcLik(currRegion, currRecomb1, currRecomb2);

	return -lik;
}




